from .generic_elongation import GenericElongationGroove


class FlatGroove(GenericElongationGroove):
    """Represents a box shaped groove."""

    def __init__(
            self,
            width: float,
    ):
        """
        :param width: width of the forming zone
        :type width: float
        """

        super().__init__(usable_width=width)

    @property
    def types(self) -> '("flat",)':
        return "flat",



import math

import numpy as np
from shapely.geometry import Polygon
from shapely.ops import clip_by_rect

from ..roll_pass import RollPass


@RollPass.InProfile.x
def entry_point(self: RollPass.InProfile):
    return -self.roll_pass().roll.contact_length


@RollPass.OutProfile.x
def exit_point(self: RollPass.OutProfile):
    return 0


@RollPass.OutProfile.strain
def strain(self: RollPass.OutProfile):
    return self.roll_pass().in_profile.strain + self.roll_pass().strain


@RollPass.OutProfile.width
def width(self: RollPass.OutProfile):
    return self.roll_pass().roll.groove.usable_width


@RollPass.OutProfile.width
def width(self: RollPass.OutProfile):
    if self.has_set_or_cached("cross_section") and self.has_set_or_cached("equivalent_width"):
        return self.equivalent_width ** 2 / self.cross_section.area * self.height


@RollPass.OutProfile.length
def length(self: RollPass.OutProfile):
    return self.roll_pass().elongation * self.roll_pass().in_profile.length


@RollPass.OutProfile.filling_ratio
def filling_ratio(self: RollPass.OutProfile):
    return self.width / self.roll_pass().roll.groove.usable_width


@RollPass.OutProfile.cross_section
def cross_section(self: RollPass.OutProfile) -> Polygon:
    poly = Polygon(np.concatenate([
        self.roll_pass().upper_contour_line.coords,
        self.roll_pass().lower_contour_line.coords
    ]))

    if (
            # one percent tolerance to bypass discretization issues
            - self.width / 2 < poly.bounds[0] * 1.01
            or self.width / 2 > poly.bounds[2] * 1.01
    ):
        raise ValueError("Profile's width can not be larger than its contour lines."
                         "May be caused by critical overfilling.")

    return clip_by_rect(poly, -self.width / 2, -math.inf, self.width / 2, math.inf)


@RollPass.OutProfile.types
def types(self: RollPass.OutProfile):
    return self.roll_pass().types



import numpy as np

from ..generic_elongation import GenericElongationGroove


class FlatOvalGroove(GenericElongationGroove):
    """Represent an oval shaped groove with a flat ground."""

    def __init__(self, r1: float, r2: float, depth: float, usable_width: float):
        """
        :param r1: radius of the first edge
        :type r1: float
        :param r2: radius of the second edge
        :type r2: float
        :param depth: depth of the groove
        :type depth: float
        :param usable_width:  ground width excluding influence of radii
        :type usable_width: float
        """
        alpha = np.arccos(1 - depth / (r1 + r2))
        even_ground_width = usable_width - 2 * (r1 * np.sin(alpha) + r2 * np.sin(alpha) - r1 * np.tan(alpha / 2))

        super().__init__(usable_width=usable_width, depth=depth, r1=r1, r2=r2, alpha1=alpha, alpha2=alpha,
                         even_ground_width=even_ground_width)

    @property
    def types(self) -> '("oval", "flat_oval")':
        return "oval", "flat_oval"



VERSION = "2.0.0b1-2"

from . import shapes as _

from .grooves import BoxGroove, ConstrictedBoxGroove, SquareGroove, DiamondGroove, RoundGroove, FalseRoundGroove, \
    CircularOvalGroove, FlatOvalGroove, SwedishOvalGroove, ConstrictedSwedishOvalGroove, Oval3RadiiGroove, \
    Oval3RadiiFlankedGroove, SplineGroove, GenericElongationGroove, FlatGroove
from .transport import Transport
from .roll_pass import RollPass
from .unit import Unit
from .roll import Roll
from .profile import Profile
from .rotator import Rotator
from .sequence import PassSequence
from .hooks import Hook, HookHost, HookFunction, root_hooks

root_hooks.update({
    RollPass.roll_force,
    RollPass.Roll.roll_torque,
    Unit.OutProfile.cross_section,
    Unit.OutProfile.strain,
    Unit.OutProfile.length,
    Unit.OutProfile.types,
    PassSequence.total_elongation,
})



from .base import GrooveBase
from .spline import SplineGroove
from .generic_elongation import GenericElongationGroove
from .boxes import BoxGroove, ConstrictedBoxGroove
from .diamonds import DiamondGroove, SquareGroove
from .ovals import CircularOvalGroove, FlatOvalGroove, SwedishOvalGroove, ConstrictedSwedishOvalGroove, \
    Oval3RadiiGroove, Oval3RadiiFlankedGroove
from .rounds import RoundGroove, FalseRoundGroove
from .flat import FlatGroove



from abc import ABC, abstractmethod
from typing import Union, Tuple

import numpy as np
from shapely.geometry import Polygon, LineString


class GrooveBase(ABC):
    """Abstract base class for all grooves."""

    @property
    @abstractmethod
    def types(self) -> Tuple[str, ...]:
        """A tuple of keywords to specify the types of this groove."""
        raise NotImplemented

    @property
    @abstractmethod
    def cross_section(self) -> Polygon:
        """A polygon representing the cross-section of this groove limited by the contour line and y=0."""
        raise NotImplemented

    @property
    @abstractmethod
    def usable_width(self) -> float:
        """The usable width of the groove, meaning the width of ideal filling."""
        raise NotImplemented

    @property
    @abstractmethod
    def depth(self) -> float:
        """The maximum depth of the groove."""
        raise NotImplemented

    @property
    @abstractmethod
    def contour_line(self) -> LineString:
        """A line representing the geometry of the groove contour."""
        raise NotImplemented

    @abstractmethod
    def local_depth(self, z: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Function of the local groove depth in dependence on the z-coordinate."""
        raise NotImplemented

    def __str__(self):
        return f"{type(self).__name__} {self.usable_width:.4g} x {self.depth:.4g} ({', '.join(self.types)})"



import logging
import weakref
from typing import Optional, Sequence

import numpy as np

from ..hooks import HookHost, Hook
from ..profile import Profile as BaseProfile


class Unit(HookHost):
    """Base class for units."""

    max_iteration_count = Hook[int]()
    """Count of maximum solution loop iterations before aborting."""

    iteration_precision = Hook[float]()
    """Precision of iteration break in solution loop."""

    length = Hook[float]()
    """The length of the unit (spacial extent in rolling direction)."""

    duration = Hook[float]()
    """Time needed to pass the unit (temporal extent)."""

    velocity = Hook[float]()
    """Mean velocity of material flow."""

    volume = Hook[float]()
    """Volume of workpiece material within the unit."""

    def __init__(self, label: str):
        super().__init__()
        self._log = logging.getLogger(__name__)
        self.label = label
        """Label for human identification."""

        self._subunits: Optional[Unit._SubUnitsList] = self._SubUnitsList(self, [])

        self.parent = None
        """Weak reference to the parent unit, if applicable."""

        self.in_profile = None
        """The state of the incoming profile."""

        self.out_profile = None
        """The state of the outgoing profile."""

    def __str__(self):
        if self.label:
            return type(self).__qualname__ + f" '{self.label}'"
        return type(self).__qualname__

    def init_solve(self, in_profile: BaseProfile):
        """
        Method called by the standard :py:meth:`solve` implementation to init
        the specialized profile instances for in and out.

        :param in_profile: the incoming state passed to :py:meth:`solve`
        """
        self.in_profile = self.InProfile(self, in_profile)
        self.out_profile = self.OutProfile(self, in_profile)

    def get_root_hook_results(self):
        in_profile_results = self.in_profile.evaluate_and_set_hooks()
        self_results = self.evaluate_and_set_hooks()
        out_profile_results = self.out_profile.evaluate_and_set_hooks()

        return np.concatenate([in_profile_results, self_results, out_profile_results], axis=0)

    def clear_hook_cache(self):
        self.in_profile.clear_hook_cache()
        super().clear_hook_cache()
        self.out_profile.clear_hook_cache()

    def _solve_subunits(self):
        last_profile = self.in_profile
        for u in self._subunits:
            try:
                last_profile = u.solve(last_profile)
            except Exception as e:
                raise RuntimeError(f"Solution of sub units failed at unit {u}.") from e

    def solve(self, in_profile: BaseProfile) -> BaseProfile:
        """
        Solve the workpiece evolution within this unit based on the incoming state specified by in_profile.
        Iterates several times calling :py:meth:`evaluate` until the return of :py:meth:`evaluate` does not change anymore.

        :param in_profile: The incoming state profile
        :return: The outgoing state profile.
        """
        self._log.info(f"Started solving of {self}.")
        self.init_solve(in_profile)
        old_values = np.nan

        for i in range(1, self.max_iteration_count):
            self.clear_hook_cache()
            self._solve_subunits()
            current_values = self.get_root_hook_results()

            if np.all(np.abs(current_values - old_values) <= np.abs(old_values) * self.iteration_precision):
                self._log.info(f"Finished solving of {self} after {i} iterations.")
                break

            old_values = current_values

        else:
            self._log.warning(
                f"Solution iteration of {self} exceeded the maximum iteration count of {self.max_iteration_count}."
                f" Continuing anyway.")

        result = BaseProfile(**{
            k: v for k, v in self.out_profile.__dict__.items()
            if not k.startswith("_")
        })
        return result

    class Profile(BaseProfile):
        """Represents a profile in context of a unit."""

        def __init__(self, unit: 'Unit', template: BaseProfile):
            kwargs = dict(
                e for e in template.__dict__.items()
                if not e[0].startswith("_")
            )
            super().__init__(**kwargs)
            self.unit = weakref.ref(unit)

    class InProfile(Profile):
        """Represents an incoming profile of a unit."""

    class OutProfile(Profile):
        """Represents an outgoing profile of a unit."""

    class _SubUnitsList(list):
        """Specialized list for holding the units of a pass sequence."""

        def __init__(self, owner: 'Unit', units: Sequence['Unit']):
            super().__init__(units)
            self.owner = weakref.ref(owner)
            for u in self:
                u.parent = weakref.ref(owner)

        # noinspection PyProtectedMember
        def _repr_html_(self):
            return "<br/>".join(v._repr_html_() for v in self)

    @property
    def subunits(self):
        return self._subunits



from typing import Iterable

import numpy as np
from shapely.ops import linemerge
import logging

from pyroll.core.repr import ReprMixin
from shapely import MultiLineString, Point, Polygon, LineString


# noinspection PyAbstractClass
class PatchedPolygon(Polygon):
    @property
    def height(self) -> float:
        """Computes the height of the bounding box."""
        return self.bounds[3] - self.bounds[1]

    @property
    def width(self) -> float:
        """Computes the width of the bounding box."""
        return self.bounds[2] - self.bounds[0]

    @property
    def perimeter(self) -> float:
        """Get the perimeter of the Polygon (alias of ``Polygon.length``)."""
        return self.length

    @property
    def __attrs__(self):
        return {
            "width": self.width,
            "height": self.height,
            "perimeter": self.perimeter,
            "area": self.area,
        }


Polygon.height = PatchedPolygon.height
Polygon.width = PatchedPolygon.width
Polygon.perimeter = PatchedPolygon.perimeter
Polygon.__attrs__ = PatchedPolygon.__attrs__
Polygon.__str__ = ReprMixin.__str__
Polygon.__repr__ = ReprMixin.__repr__
Polygon._repr_html_ = ReprMixin._repr_html_
Polygon._repr_pretty_ = ReprMixin._repr_pretty_

_RECTANGLE_CORNERS = np.asarray([
    (-0.5, -0.5),
    (0.5, -0.5),
    (0.5, 0.5),
    (-0.5, 0.5)
])


def rectangle(width: float, height: float):
    """
    Creates an instance of :py:class:`Polygon` with rectangular shape from height and width aligned to the axes.

    :param width: the width of the rectangle
    :param height: the height of the rectangle
    """

    try:
        width = float(width)
        height = float(height)

    except TypeError as e:
        raise TypeError("width and height must be convertible to float") from e

    points = _RECTANGLE_CORNERS * (width, height)
    rect = Polygon(points)

    return rect


class PatchedLineString(LineString):
    @property
    def depth(self) -> float:
        """Computes the height of the bounding box."""
        return self.bounds[3] - self.bounds[1]

    @property
    def width(self) -> float:
        """Computes the width of the bounding box."""
        return self.bounds[2] - self.bounds[0]

    @property
    def __attrs__(self):
        return {
            "width": self.width,
            "depth": self.depth,
            "length": self.length,
        }


LineString.depth = PatchedLineString.depth
LineString.width = PatchedLineString.width
LineString.__attrs__ = PatchedLineString.__attrs__
LineString.__str__ = ReprMixin.__str__
LineString.__repr__ = ReprMixin.__repr__
LineString._repr_html_ = ReprMixin._repr_html_
LineString._repr_pretty_ = ReprMixin._repr_pretty_


def linemerge_if_multi(lines):
    if isinstance(lines, MultiLineString) or isinstance(lines, Iterable):
        merged_line = linemerge(lines)

        if isinstance(merged_line, MultiLineString):

            logging.getLogger(__name__).warning(
                "Discontinuous LineStrings for profile separation. Doing point wise comparison to resolve."
            )

            line_lengths = [line.length for line in merged_line.geoms]
            index_of_continuous_line = np.argmax(line_lengths)
            continuous_linestring = merged_line[index_of_continuous_line]

            discontinuous_lines = list(lines)
            discontinuous_lines.pop(index_of_continuous_line)

            points_representation_of_discontinuous_lines = [Point(p) for l in discontinuous_lines for p in l.coords]

            distances_of_points_to_continuous_line = [continuous_linestring.distance(p) for p in
                                                      points_representation_of_discontinuous_lines]

            if max(distances_of_points_to_continuous_line) <= line_lengths[index_of_continuous_line] / 1000:
                return continuous_linestring

            else:
                logging.getLogger(__name__).error(
                    f"Could not resolve discontinuous LineStrings for profile separation."
                )
                raise RuntimeError("Could not resolve discontinuous LineStrings for profile separation.")

        return merged_line

    return lines



import logging
from ..unit import Unit


class DiskElement(Unit):
    """
    Represents a disk element unit, a unit used to subdivide other units in rolling direction.
    The basic disk element is not intended to be used directly.
    Use the specialized disk elements of the respective parent unit instead.
    """

    def __init__(
            self,
            parent: 'Unit',
            index: int,
            **kwargs
    ):
        """
        :param label: label for human identification
        :param kwargs: additional hook values as keyword arguments to set explicitly
        """

        super().__init__(label=f"{parent}[{index}]")
        self.__dict__.update(kwargs)
        self._log = logging.getLogger(__name__)

    class Profile(Unit.Profile):
        """Represents a profile in context of a disk element unit."""

        def disk_element(self) -> 'DiskElement':
            return self.unit()

    class InProfile(Profile, Unit.InProfile):
        """Represents an incoming profile of a disk element unit."""

    class OutProfile(Profile, Unit.OutProfile):
        """Represents an outgoing profile of a disk element unit."""



from typing import Optional

import numpy as np

from ..generic_elongation import GenericElongationGroove


class SwedishOvalGroove(GenericElongationGroove):
    """Represents a hexagonal shaped groove that is used like an oval groove (swedish oval)."""

    def __init__(
            self,
            r1: float,
            r2: float,
            depth: float,
            ground_width: Optional[float] = None,
            usable_width: Optional[float] = None,
            flank_angle: Optional[float] = None
    ):
        """
        Exactly two of ground_width, usable_width and flank_angle must be given.

        :param r1: radius of the first edge
        :type r1: float
        :param r2: radius of the second edge
        :type r2: float
        :param depth: depth of the groove
        :type depth: float
        :param ground_width: width of the groove from intersection between two flanks and ground width
        :type ground_width: float
        :param usable_width:  ground width excluding influence of radii
        :type usable_width: float
        :param flank_angle: angle of the flanks
        :type flank_angle: float
        :raises ValueError:  if not exactly two of ground_width, usable_width and flank_angle are given
        """

        if ground_width and usable_width and not flank_angle:
            flank_angle = np.arctan(depth / (usable_width - ground_width) * 2)
        elif usable_width and flank_angle and not ground_width:
            ground_width = usable_width - 2 * depth / np.tan(flank_angle)
        elif ground_width and flank_angle and not usable_width:
            usable_width = ground_width + 2 * depth / np.tan(flank_angle)
        else:
            raise ValueError(
                "Exactly two of the following arguments must be given: ground_width, usable_width, flank_angle."
            )

        even_ground_width = ground_width - 2 * r2 * np.tan(flank_angle / 2)

        super().__init__(usable_width=usable_width, depth=depth, r1=r1, r2=r2, alpha1=flank_angle, alpha2=flank_angle,
                         even_ground_width=even_ground_width)

    @property
    def types(self) -> '("oval", "swedish_oval")':
        return "oval", "swedish_oval"



import logging
from typing import Iterable, Union

import numpy as np
from shapely.geometry import LineString

from ..grooves import GrooveBase
from ..hooks import HookHost, Hook


class Roll(HookHost):
    """Represents a roll."""

    nominal_radius = Hook[float]()
    """Nominal radius."""

    working_radius = Hook[float]()
    """Effective working radius (equivalent flat rolling)."""

    min_radius = Hook[float]()
    """Minimum radius (at lowest point of groove)."""

    max_radius = Hook[float]()
    """Maximum radius (at highest point of groove)."""

    rotational_frequency = Hook[float]()
    """Rotational frequency (revolutions per time)."""

    surface_velocity = Hook[float]()
    """Tangential velocity of the outer roll surface (at nominal radius)."""

    contour_line = Hook[LineString]()
    """Contour line in the z-y-plane."""

    roll_torque = Hook[float]()
    """Roll torque of single roll."""

    roll_power = Hook[float]()
    """Roll power of single roll."""

    contact_length = Hook[float]()
    """Length of the longest contact arc between roll and workpiece."""

    contact_area = Hook[float]()
    """Total area of contact between roll and workpiece."""

    center = Hook[np.ndarray]()
    """Center point of the roll as 1D array."""

    temperature = Hook[float]()
    """Mean temperature."""

    surface_temperature = Hook[float]()
    """Effective surface temperature."""

    yield_strength = Hook[float]()
    """Yield strength of the roll material."""

    elastic_modulus = Hook[float]()
    """Elastic modulus of the roll material."""

    poissons_ratio = Hook[float]()
    """Poisson's ratio of the roll material."""

    thermal_conductivity = Hook[float]()
    """Thermal conductivity of the roll material."""

    thermal_capacity = Hook[float]()
    """Thermal capacity of the roll material."""

    density = Hook[float]()
    """Density (specific weight) of the roll material."""

    material = Hook[Union[str, Iterable[str]]]()
    """String or sequence of strings classifying the material of the roll.
    Can be used by material databases to retrieve respective data."""

    def __init__(
            self,
            groove: GrooveBase,
            **kwargs):
        """
        :param groove: the groove object defining the shape of the roll's surface
        :param kwargs: additional hook values as keyword arguments to set explicitly
        """
        self.__dict__.update(kwargs)

        super().__init__()

        self.groove = groove
        """The groove object defining the shape of the roll's surface."""

        self._log = logging.getLogger(__name__)



import numpy as np

from .roll import Roll


@Roll.working_radius
def working_radius(self: Roll):
    return self.nominal_radius - self.contour_line.centroid.y


@Roll.min_radius
def min_radius(self: Roll):
    return self.max_radius - self.contour_line.bounds[3]


@Roll.max_radius
def max_radius(self: Roll):
    return self.nominal_radius


@Roll.contour_line
def contour_line(self: Roll):
    return self.groove.contour_line


@Roll.roll_power
def roll_power(self: Roll):
    return self.roll_torque * self.rotational_frequency


@Roll.surface_velocity
def surface_velocity(self: Roll):
    return self.rotational_frequency * self.nominal_radius


@Roll.rotational_frequency
def rotational_frequency(self: Roll):
    return self.surface_velocity / self.nominal_radius


from .rotator import Rotator
from shapely.affinity import rotate


@Rotator.OutProfile.cross_section
def rotated_cross_section(self: Rotator.OutProfile):
    r = self.rotator()
    return rotate(r.in_profile.cross_section, angle=r.rotation, origin=(0, 0))



from .disk_element import DiskElement
from .disked_unit import DiskedUnit


@DiskElement.OutProfile.x
def out_x(self: DiskElement.OutProfile):
    return self.disk_element().in_profile.x + self.disk_element().length


@DiskElement.OutProfile.t
def out_t(self: DiskElement.OutProfile):
    return self.disk_element().in_profile.t + self.disk_element().duration


@DiskedUnit.disk_element_count
def no_disks(self: DiskedUnit):
    return 0



from .transport import Transport


@Transport.OutProfile.strain
def out_strain(self: Transport.OutProfile):
    """Assume total recrystallization during transport."""
    return 0


@Transport.OutProfile.velocity
def out_velocity(self: Transport.OutProfile):
    return self.transport().velocity


@Transport.velocity
def velocity(self: Transport):
    return self.in_profile.velocity


@Transport.duration
def duration(self: Transport):
    return self.length / self.velocity



import numpy as np

from ..roll_pass import RollPass
from ...grooves import GenericElongationGroove


@RollPass.mean_flow_stress
def mean_flow_stress(self: RollPass):
    return (self.in_profile.flow_stress + 2 * self.out_profile.flow_stress) / 3


@RollPass.roll_force
def roll_force(self: RollPass):
    return self.mean_flow_stress * self.roll.contact_area


@RollPass.tip_width
def tip_width(self):
    if isinstance(self.roll.groove, GenericElongationGroove):
        return self.roll.groove.usable_width + self.gap / 2 / np.tan(self.roll.groove.alpha1)


@RollPass.height
def height(self):
    return self.gap + 2 * self.roll.groove.depth


@RollPass.volume
def volume(self: RollPass):
    return (self.in_profile.cross_section.area + 2 * self.out_profile.cross_section.area
            ) / 3 * self.roll.contact_length


@RollPass.velocity
def velocity(self: RollPass):
    return self.roll.working_radius * self.roll.rotational_frequency


@RollPass.draught
def draught(self: RollPass):
    return self.out_profile.equivalent_rectangle.height / self.in_profile.equivalent_rectangle.height


@RollPass.spread
def spread(self: RollPass):
    return self.out_profile.equivalent_rectangle.width / self.in_profile.equivalent_rectangle.width


@RollPass.elongation
def elongation(self: RollPass):
    return self.in_profile.cross_section.area / self.out_profile.cross_section.area


@RollPass.log_draught
def log_draught(self: RollPass):
    return np.log(self.draught)


@RollPass.log_spread
def log_spread(self: RollPass):
    return np.log(self.spread)


@RollPass.log_elongation
def log_elongation(self: RollPass):
    return np.log(self.elongation)


@RollPass.abs_draught
def abs_draught(self: RollPass):
    return self.out_profile.equivalent_rectangle.height - self.in_profile.equivalent_rectangle.height


@RollPass.abs_spread
def abs_spread(self: RollPass):
    return self.out_profile.equivalent_rectangle.width - self.in_profile.equivalent_rectangle.width


@RollPass.abs_elongation
def abs_elongation(self: RollPass):
    return self.out_profile.length - self.in_profile.length


@RollPass.rel_draught
def rel_draught(self: RollPass):
    return self.abs_draught / self.in_profile.equivalent_rectangle.heigth


@RollPass.rel_spread
def rel_spread(self: RollPass):
    return self.abs_spread / self.in_profile.equivalent_rectangle.width


@RollPass.rel_elongation
def rel_elongation(self: RollPass):
    return self.abs_elongation / self.in_profile.length


@RollPass.strain
def strain(self: RollPass):
    return self.log_elongation


@RollPass.strain_rate
def strain_rate(self: RollPass):
    return self.velocity / self.roll.contact_length * self.draught



from .circular_oval import CircularOvalGroove
from .flat_oval import FlatOvalGroove
from .swedish_oval import SwedishOvalGroove
from .constricted_swedish_oval import ConstrictedSwedishOvalGroove
from .oval_3radii import Oval3RadiiGroove
from .oval_3radii_flanked import Oval3RadiiFlankedGroove



from ..roll_pass import RollPass


@RollPass.in_profile_rotation
def diamonds(self: RollPass):
    if "diamond" in self.in_profile.types and "diamond" in self.types:
        return 90


@RollPass.in_profile_rotation
def oval_round(self: RollPass):
    if "oval" in self.in_profile.types and "round" in self.types:
        return 90


@RollPass.in_profile_rotation
def round_oval(self: RollPass):
    if "round" in self.in_profile.types and "oval" in self.types:
        return 90


@RollPass.in_profile_rotation
def oval_square(self: RollPass):
    if "oval" in self.in_profile.types and "square" in self.types:
        return 90


@RollPass.in_profile_rotation
def square_oval(self: RollPass):
    if "square" in self.in_profile.types and "oval" in self.types:
        return 45


@RollPass.in_profile_rotation
def box_box(self: RollPass):
    if "box" in self.in_profile.types and "box" in self.types:
        return 90


@RollPass.in_profile_rotation
def box_diamond(self: RollPass):
    if "box" in self.in_profile.types and "diamond" in self.types:
        return 45


@RollPass.in_profile_rotation
def box_oval(self: RollPass):
    if "box" in self.in_profile.types and "oval" in self.types:
        return 90


@RollPass.in_profile_rotation
def round_flat(self: RollPass):
    if "round" in self.in_profile.types and "flat" in self.types:
        return 0



import logging
import math
from typing import Optional, Tuple, Iterable, Union

import numpy as np
from shapely.affinity import translate
from shapely.geometry import Point, LinearRing, Polygon, LineString
from shapely.ops import clip_by_rect, unary_union

from ..grooves import GrooveBase
from ..hooks import HookHost, Hook

_log = logging.getLogger(__name__)


class Profile(HookHost):
    """Represents a profile aka a workpiece state."""

    cross_section = Hook[Polygon]()
    """Shape of the profile's cross-section."""

    types = Hook[Tuple[str, ...]]()
    """Classifiers of the profile's shape's type."""

    x = Hook[float]()
    """Spacial coordinate in rolling direction."""

    t = Hook[float]()
    """Temporal coordinate."""

    velocity = Hook[float]()
    """Mean material flow velocity."""

    height = Hook[float]()
    """Maximum height (y-direction)."""

    width = Hook[float]()
    """Maximum width (z-direction)."""

    length = Hook[float]()
    """Length of the workpiece (x-direction)."""

    equivalent_rectangle = Hook[Polygon]()
    """Equivalent rectangle geometry for use with equivalent flat pass models."""

    equivalent_height = Hook[float]()
    """Height of the equivalent rectangle."""

    equivalent_width = Hook[float]()
    """Width of the equivalent rectangle."""

    temperature = Hook[float]()
    """Mean temperature of the profile cross-section."""

    surface_temperature = Hook[float]()
    """Mean temperature of the profile surface."""

    core_temperature = Hook[float]()
    """Temperature of the profile core."""

    strain = Hook[float]()
    """Mean equivalent strain of the profile cross-section."""

    flow_stress = Hook[float]()
    """Mean flow stress of the profile material."""

    elastic_modulus = Hook[float]()
    """Mean elastic modulus of the profile material."""

    poissons_ratio = Hook[float]()
    """Mean Poisson's ratio of the profile material."""

    thermal_conductivity = Hook[float]()
    """Mean thermal conductivity of the profile material."""

    thermal_capacity = Hook[float]()
    """Mean thermal capacity of the profile material."""

    density = Hook[float]()
    """Mean density (specific weight) of the profile material."""

    material = Hook[Union[str, Iterable[str]]]()
    """String or sequence of strings classifying the material of the profile.
    Can be used by material databases to retrieve respective data."""

    def __init__(self, **kwargs):
        """Using the ``__init__`` is not recommended, use one of the factory class methods instead."""
        self.__dict__.update(kwargs)
        super().__init__()

    @classmethod
    def from_groove(
            cls,
            groove: GrooveBase,
            width: Optional[float] = None,
            filling: Optional[float] = None,
            height: Optional[float] = None,
            gap: Optional[float] = None,
            **kwargs
    ) -> 'Profile':
        """
        Create a profile instance based on a given groove.
        The dimensioning of the profile is determined by the parameters ``width``, ``filling``, ``height`` and ``gap``.
        Give exactly one of ``width`` and ``filling``.
        Give exactly one of ``height`` and ``gap``.

        :param groove: the groove the profile should be created from
        :param width: the width of the resulting profile, must be > 0
        :param filling: the filling ratio of the groove, must be > 0
        :param height: the height of the profile, must be > 0
        :param gap: the gap between the groove contours (roll gap), must be >= 0
        :param kwargs: additional keyword arguments to be passed to the Profile constructor
        :raises TypeError: on invalid argument combinations
        :raises ValueError: if arguments are out of range
        """

        if width is not None and filling is None:
            filling = width / groove.usable_width
        elif filling is not None and width is None:
            width = filling * groove.usable_width
        else:
            raise TypeError("either 'width' or 'filling' must be given")

        if height is not None and gap is None:
            gap = height - 2 * groove.depth
        elif gap is not None and height is None:
            height = gap + 2 * groove.depth
        else:
            raise TypeError("either 'gap' or 'height' must be given")

        if (
                filling <= 0
                or width <= 0
                or height <= 0
                or gap < 0

        ):
            raise ValueError("argument value(s) out of range")

        if filling > 1:
            _log.warning("Encountered overfilled groove in profile construction.")

        upper_contour_line = translate(groove.contour_line, yoff=gap / 2)
        lower_contour_line = translate(groove.contour_line, yoff=-gap / 2)

        poly = Polygon(np.concatenate([
            upper_contour_line.coords,
            lower_contour_line.coords
        ]))

        if (
                # one percent tolerance to bypass discretization issues
                - width / 2 < poly.bounds[0] * 1.01
                or width / 2 > poly.bounds[2] * 1.01
        ):
            raise ValueError("Profile's width can not be larger than its contour lines."
                             "May be caused by critical overfilling.")

        polygon = clip_by_rect(poly, -width / 2, -math.inf, width / 2, math.inf)

        return cls(
            cross_section=polygon,
            types=groove.types,
            **kwargs
        )

    @classmethod
    def round(
            cls,
            radius: Optional[float] = None,
            diameter: Optional[float] = None,
            **kwargs
    ) -> 'Profile':
        """
        Creates a round shaped profile (a real circle round, without imperfections of round grooves).
        Give exactly one of ``radius`` and ``diameter``.

        :param radius: the radius of the round profile, must be > 0
        :param diameter: the diameter of the round profile, must be > 0
        :param kwargs: additional keyword arguments to be passed to the Profile constructor
        :raises TypeError: on invalid argument combinations
        :raises ValueError: if arguments are out of range
        """

        if radius is not None and diameter is None:
            diameter = 2 * radius
        elif diameter is not None and radius is None:
            radius = diameter / 2
        else:
            raise TypeError("either 'radius' or 'diameter' must be given")

        if radius <= 0:
            raise ValueError("argument value(s) out of range")

        center = Point((0, 0))
        circle = center.buffer(radius)

        return cls(
            cross_section=circle,
            types=["round"],
            **kwargs
        )

    @classmethod
    def square(
            cls,
            side: Optional[float] = None,
            diagonal: Optional[float] = None,
            corner_radius: float = 0,
            **kwargs
    ) -> 'Profile':
        """
        Creates a square shaped profile (a real square with rounded corners, without imperfections of square grooves).
        A square is oriented to stand on its corner, use :py:meth:`box` to create a side standing one.
        Give exactly one of ``side`` and ``diagonal``.

        :param side: the side length of the square profile, must be > 0
        :param diagonal: the diagonal's length of the square profile, must be > 0.
            Note, that the diagonal is measured at the tips, as if the corner radii were not present
            for consistency with :py:meth:`box`.
        :param corner_radius: the radius of the square's corners, must be >= 0 and <= side / 2
        :param kwargs: additional keyword arguments to be passed to the Profile constructor
        :raises TypeError: on invalid argument combinations
        :raises ValueError: if arguments are out of range
        """

        if side is not None and diagonal is None:
            diagonal = np.sqrt(2) * side
        elif diagonal is not None and side is None:
            side = diagonal / np.sqrt(2)
        else:
            raise TypeError("either 'side' or 'diagonal' must be given")

        if (
                side <= 0
                or corner_radius < 0
                or corner_radius > side / 2
        ):
            raise ValueError("argument value(s) out of range")

        line = LinearRing(np.array([(1, 0), (0, 1), (-1, 0), (0, -1)]) * (diagonal / 2 - corner_radius * np.sqrt(2)))
        polygon = Polygon(line)
        polygon = polygon.buffer(corner_radius)

        return cls(
            cross_section=polygon,
            types=["square", "diamond"],
            **kwargs
        )

    @classmethod
    def box(
            cls,
            height: float,
            width: float,
            corner_radius: float = 0,
            **kwargs
    ) -> 'Profile':
        """
        Creates a box shaped profile (a real rectangular shape with rounded corners,
        without imperfections of box grooves).
        A box is oriented to stand on its side, use :py:meth:`square` to create a corner standing square.

        :param height: the height of the box profile, must be > 0
        :param width: the width of the box profile, must be > 0
        :param corner_radius: the radius of the square's corners, must be >= 0, <= height / 2 and <= width / 2
        :param kwargs: additional keyword arguments to be passed to the Profile constructor
        :raises ValueError: if arguments are out of range
        """

        if (
                height <= 0
                or width <= 0
                or corner_radius < 0
                or corner_radius > height / 2
                or corner_radius > width / 2
        ):
            raise ValueError("argument value(s) out of range")

        line = LinearRing(np.array([(1, -1), (1, 1), (-1, 1), (-1, -1)])
                          * (width / 2 - corner_radius, height / 2 - corner_radius))
        polygon = Polygon(line)
        polygon = polygon.buffer(corner_radius)

        return cls(
            cross_section=polygon,
            types=["box"],
            **kwargs
        )

    @classmethod
    def diamond(
            cls,
            height: float,
            width: float,
            corner_radius: float = 0,
            **kwargs
    ) -> 'Profile':
        """
        Creates a diamond shaped profile (a real diamond shape with rounded corners,
        without imperfections of diamond grooves).
        A diamond is oriented to stand on its corner.

        :param height: the height of the diamond profile, must be > 0
        :param width: the width of the diamond profile, must be > 0
        :param corner_radius: the radius of the diamonds's corners, must be >= 0, <= height / 2 and <= width / 2
        :param kwargs: additional keyword arguments to be passed to the Profile constructor
        :raises ValueError: if arguments are out of range
        """

        if (
                height <= 0
                or width <= 0
                or corner_radius < 0
                or corner_radius > height / 2
                or corner_radius > width / 2
        ):
            raise ValueError("argument value(s) out of range")

        line = LinearRing(np.array([(1, 0), (0, 1), (-1, 0), (0, -1)])
                          * (width / 2 - corner_radius, height / 2 - corner_radius))
        polygon = Polygon(line)
        polygon = polygon.buffer(corner_radius)

        return cls(
            cross_section=polygon,
            types=["diamond"],
            **kwargs
        )

    def local_height(self, z: float) -> float:
        coords = np.array([(1, -1), (1, 1)]) * (z, self.height)

        vline = LineString(
            coords
        )

        intersection = vline.intersection(self.cross_section)

        return intersection.length

    def local_width(self, y: float) -> float:
        coords = np.array([(-1, 1), (1, 1)]) * (self.width, y)

        hline = LineString(
            coords
        )

        intersection = hline.intersection(self.cross_section)

        return intersection.length



from .unit import Unit

for h in Unit.OutProfile.__hooks__:
    @getattr(Unit.OutProfile, h)(trylast=True)
    def copy_from_in_profile(self: Unit.OutProfile, hook=h):
        return getattr(self.unit().in_profile, hook, None)


    @getattr(Unit.OutProfile, h)
    def copy_from_last_subunit(self: Unit.OutProfile, hook=h):
        if self.unit().subunits:
            return getattr(self.unit().subunits[-1].out_profile, hook, None)


@Unit.iteration_precision
def iteration_precision(self: Unit):
    return 1e-2


@Unit.max_iteration_count
def max_iteration_count(self: Unit):
    return 100


@Unit.volume
def volume(self: Unit):
    return (self.in_profile.cross_section.area + self.out_profile.cross_section.area) / 2 * self.length


@Unit.length
def length(self: Unit):
    if self.has_set_or_cached("duration"):
        return self.velocity * self.duration


@Unit.duration
def duration(self: Unit):
    if self.has_set_or_cached("length"):
        return self.length / self.velocity


@Unit.velocity
def velocity(self: Unit):
    if self.in_profile.has_set_or_cached("velocity"):
        return self.in_profile.velocity



import logging
import weakref

from ..hooks import Hook
from ..unit import Unit
from ..profile import Profile as BaseProfile


class Rotator(Unit):
    """Represents a unit rotating a profile around the rolling axis (mostly for feeding into next roll pass)."""

    rotation = Hook[float]()
    """Rotation applied to the profile in ° (degree)."""

    def __init__(
            self,
            label: str = "",
            **kwargs
    ):
        """
        :param label: label for human identification
        :param kwargs: additional hook values as keyword arguments to set explicitly
        """

        super().__init__(label)
        self.__dict__.update(kwargs)
        self._log = logging.getLogger(__name__)

    class Profile(Unit.Profile):
        """Represents a profile in context of a rotator."""

        def __init__(self, rotator: 'Rotator', template: BaseProfile):
            super().__init__(rotator, template)
            self.rotator = weakref.ref(rotator)

    class InProfile(Profile, Unit.InProfile):
        """Represents an incoming profile of a rotator."""

    class OutProfile(Profile, Unit.OutProfile):
        """Represents an outgoing profile of a rotator."""



import numpy as np

from ..roll_pass import RollPass


@RollPass.Roll.roll_torque
def roll_torque(self: RollPass.Roll):
    return self.roll_pass().roll_force * self.contact_length * 0.5


@RollPass.Roll.contact_length
def contact_length(self: RollPass.Roll):
    """
    Contact length between rolls and stock calculated using Siebel's approach
    """
    height_change = self.roll_pass().in_profile.height - self.roll_pass().height
    return np.sqrt(self.min_radius * height_change - height_change ** 2 / 4)


@RollPass.Roll.contact_area
def contact_area(self: RollPass.Roll):
    """
    Contact area between rolls and stock calculated using Siebel's approach
    """
    return (self.roll_pass().in_profile.width + self.roll_pass().out_profile.width) / 2 * self.contact_length


@RollPass.Roll.center
def center(self: RollPass.Roll):
    return np.array([0, self.roll_pass().gap / 2 + self.nominal_radius])



from typing import Optional

import numpy as np

from .diamond import DiamondGroove


class SquareGroove(DiamondGroove):
    """Represents a square shaped groove (diamond with tip angle near 90°)."""

    def __init__(
            self,
            r1: float,
            r2: float,
            usable_width: Optional[float] = None,
            tip_depth: Optional[float] = None,
            tip_angle: Optional[float] = None
    ):
        """
        Exactly two of usable_width, tip_depth and tip_angle must be given.

        :param r1: radius of the first edge
        :type r1: float
        :param r2: radius of the second edge
        :type r2: float
        :param usable_width:  ground width excluding influence of radii
        :type usable_width: float
        :param tip_depth: depth of the tip of the groove
        :type tip_depth: float
        :param tip_angle: angle at witch the tip is formed
        :type tip_angle: float
        :raises ValueError: if not exactly two of usable_width, tip_depth and tip_angle are given
        :raises ValueError: if tip angle is <85° or >95° (no matter if given or calculated internally)
        """

        super().__init__(r1=r1, r2=r2, usable_width=usable_width, tip_depth=tip_depth, tip_angle=tip_angle)

        tip_angle_deg = 180 - 2 * self.alpha1 / np.pi * 180
        if tip_angle_deg > 95 or tip_angle_deg < 85:
            raise ValueError(
                "The tip angle of this SquareGroove significantly deviates from 90°, "
                "you should use a DiamondGroove instead."
            )

    @property
    def types(self) -> '("diamond", "square")':
        return "diamond", "square"



from typing import Optional

import numpy as np

from ..generic_elongation import GenericElongationGroove


class DiamondGroove(GenericElongationGroove):
    """Represent a diamond shaped groove."""

    def __init__(
            self,
            r1: float,
            r2: float,
            usable_width: Optional[float] = None,
            tip_depth: Optional[float] = None,
            tip_angle: Optional[float] = None
    ):
        """
        Exactly two of usable_width, tip_depth and tip_angle must be given.

        :param r1: radius of the first edge
        :type r1: float
        :param r2: radius of the second edge
        :type r2: float
        :param usable_width:  ground width excluding influence of radii
        :type usable_width: float
        :param tip_depth: depth of the tip of the groove
        :type tip_depth: float
        :param tip_angle: angle at witch the tip is formed
        :type tip_angle: float
        :raises ValueError: if not exactly two of usable_width, tip_depth and tip_angle are given
        """

        if usable_width and tip_depth and not tip_angle:
            alpha = np.arctan(tip_depth / (usable_width / 2))

        elif usable_width and tip_angle and not tip_depth:
            alpha = np.pi / 2 - tip_angle / 2
            tip_depth = usable_width / 2 * np.tan(alpha)

        elif tip_depth and tip_angle and not usable_width:
            alpha = np.pi / 2 - tip_angle / 2
            usable_width = tip_depth / np.tan(alpha) * 2
        else:
            raise ValueError(
                "Exactly two of the following arguments must be given: usable_width, tip_depth, tip_angle."
            )

        depth = tip_depth - r2 / np.cos(alpha) + r2

        super().__init__(usable_width=usable_width, depth=depth, r1=r1, r2=r2, alpha1=alpha, alpha2=alpha)

    @property
    def types(self) -> '("diamond",)':
        return "diamond",



import numpy as np

from ..generic_elongation import GenericElongationGroove


class RoundGroove(GenericElongationGroove):
    """Represents a round shaped groove."""

    def __init__(self, r1: float, r2: float, depth: float):
        """
        :param r1: radius of the first edge
        :type r1: float
        :param r2: radius of the second edge
        :type r2: float
        :param depth: depth of the groove
        :type depth: float
        """

        alpha = np.arccos(1 - depth / (r1 + r2))
        usable_width = 2 * (r1 * np.sin(alpha) + r2 * np.sin(alpha) - r1 * np.tan(alpha / 2))

        super().__init__(usable_width=usable_width, depth=depth, r1=r1, r2=r2, alpha1=alpha, alpha2=alpha)

    @property
    def types(self) -> '("round",)':
        return "round",



from typing import Union, Tuple, Iterable, Optional

import numpy as np
import scipy.interpolate
from shapely.geometry import Polygon, LineString

from pyroll.core.grooves import GrooveBase


class SplineGroove(GrooveBase):
    """Represents a groove defined by a linear spline contour."""

    def __init__(self,
                 contour_points: Union[Iterable[Tuple[float, float]], np.ndarray],
                 types: Iterable[str],
                 usable_width: Optional[float] = None,
                 ):
        """
        :param contour_points: an iterable of contour points to be used for the spline
        :param types: an interable of string keys used as type classifiers
        :param usable_width: the usable width to assume for this instance, if None, the maximum width will be used
        """
        contour_points = np.asarray(contour_points, dtype="float64")

        if contour_points.ndim != 2:
            raise ValueError("contour_points must be two-dimensional")

        if contour_points.shape[1] != 2:
            raise ValueError("contour_points must have length of 2 in second dimension")

        if not np.isclose(contour_points[0, 1], 0) or not np.isclose(contour_points[-1, 1], 0):
            raise ValueError("first and last element of contour_points should have y coordinate equal to 0")

        # strip boundary
        contour_points = contour_points[np.logical_not(
            (np.isclose(np.roll(contour_points[:, 1], 1), 0)) & (np.isclose(np.roll(contour_points[:, 1], -1), 0)))]

        # shift center to 0,0
        contour_points[:, 0] -= np.mean(contour_points[:, 0])

        half_width = contour_points[-1, 0]

        if usable_width:
            self._usable_width = usable_width
        else:
            self._usable_width = half_width * 2

        # buffer boundaries
        contour_points = np.insert(contour_points, 0, (-half_width * 1.1, 0), axis=0)
        contour_points = np.append(contour_points, [(half_width * 1.1, 0)], axis=0)

        self._contour_line = LineString(contour_points)
        self._cross_section = Polygon(self._contour_line)

        self._depth = np.max(contour_points[:, 1])

        self._local_depth = scipy.interpolate.interp1d(contour_points[:, 0], contour_points[:, 1],
                                                       fill_value="extrapolate")
        self._types = tuple(types)

    @property
    def types(self) -> Tuple[str, ...]:
        return self._types

    @property
    def cross_section(self) -> Polygon:
        return self._cross_section

    @property
    def usable_width(self) -> float:
        return self._usable_width

    @property
    def depth(self) -> float:
        return self._depth

    @property
    def contour_line(self) -> LineString:
        return self._contour_line

    def local_depth(self, z) -> Union[float, np.ndarray]:
        return self._local_depth(z)



import logging
import weakref
from typing import List

import numpy as np
from shapely.affinity import translate, rotate
from shapely.geometry import LineString, Polygon

from ..disk_element import DiskedUnit
from ..hooks import Hook
from ..profile import Profile as BaseProfile
from ..roll import Roll as BaseRoll
from ..rotator import Rotator


class RollPass(DiskedUnit):
    """Represents a roll pass."""

    in_profile_rotation = Hook[float]()
    """Rotation applied to the incoming profile in ° (degree)."""

    gap = Hook[float]()
    """Gap between the rolls (outer surface)."""

    height = Hook[float]()
    """Maximum height of the roll pass."""

    tip_width = Hook[float]()
    """Width of the intersection of the extended groove flanks (theoretical maximum filling width)."""

    roll_force = Hook[float]()
    """Vertical roll force."""

    mean_flow_stress = Hook[float]()
    """Mean value of the workpiece's flow stress within the pass."""

    spread = Hook[float]()
    """Coefficient of spread (change in width)."""

    elongation = Hook[float]()
    """Coefficient of elongation (change in length)."""

    draught = Hook[float]()
    """Coefficient of draught (change in height)."""

    log_spread = Hook[float]()
    """Log. coefficient of spread (change in width)."""

    log_elongation = Hook[float]()
    """Log. coefficient of elongation (change in length)."""

    log_draught = Hook[float]()
    """Log. coefficient of draught (change in height)."""

    abs_spread = Hook[float]()
    """Absolute spread (change in width)."""

    abs_elongation = Hook[float]()
    """Absolute elongation (change in length)."""

    abs_draught = Hook[float]()
    """Absolute draught (change in height)."""

    rel_spread = Hook[float]()
    """Relative spread (change in width)."""

    rel_elongation = Hook[float]()
    """Relative elongation (change in length)."""

    rel_draught = Hook[float]()
    """Relative draught (change in height)."""

    strain = Hook[float]()
    """Mean equivalent strain applied to the workpiece within the roll pass."""

    strain_rate = Hook[float]()
    """Mean equivalent strain rate within the roll pass."""

    def __init__(
            self,
            roll: BaseRoll,
            label: str = "",
            **kwargs
    ):
        """
        :param roll: the roll object representing the equal working rolls of the pass
        :param label: label for human identification
        :param kwargs: additional hook values as keyword arguments to set explicitly
        """

        super().__init__(label)

        self.roll = self.Roll(roll, self)
        """The working roll of this pass (equal upper and lower)."""

        self.__dict__.update(kwargs)
        self._log = logging.getLogger(__name__)

    def local_height(self, z: float) -> float:
        coords = np.array([(1, -1), (1, 1)]) * (z, self.height)

        vline = LineString(
            coords
        )

        poly = Polygon(np.concatenate([
            self.upper_contour_line.coords,
            self.lower_contour_line.coords
        ]))

        intersection = vline.intersection(poly)

        return intersection.length

    @property
    def upper_contour_line(self) -> LineString:
        """Contour line object of the upper working roll."""
        return translate(self.roll.contour_line, yoff=self.gap / 2)

    @property
    def lower_contour_line(self) -> LineString:
        """Contour line object of the lower working roll."""
        return rotate(self.upper_contour_line, angle=180, origin=(0, 0))

    @property
    def types(self):
        """A tuple of keywords to specify the shape types of this roll pass.
        Shortcut to ``self.groove.types``."""
        return self.roll.groove.types

    @property
    def disk_elements(self) -> List['RollPass.DiskElement']:
        """A list of disk elements used to subdivide this unit."""
        return list(self._subunits)

    def init_solve(self, in_profile: BaseProfile):
        super().init_solve(in_profile)

        rotator = Rotator(rotation=self.in_profile_rotation, duration=0, length=0)
        rotator.solve(in_profile)

        self.in_profile = self.InProfile(self, rotator.out_profile)
        self.out_profile = self.OutProfile(self, rotator.out_profile)

    def get_root_hook_results(self):
        super_results = super().get_root_hook_results()
        roll_results = self.roll.evaluate_and_set_hooks()

        return np.concatenate([super_results, roll_results], axis=0)

    def clear_hook_cache(self):
        super().clear_hook_cache()
        self.roll.clear_hook_cache()

    class Profile(DiskedUnit.Profile):
        """Represents a profile in context of a roll pass."""

        def __init__(self, roll_pass: 'RollPass', template: BaseProfile):
            super().__init__(roll_pass, template)
            self.roll_pass = weakref.ref(roll_pass)

    class InProfile(Profile, DiskedUnit.InProfile):
        """Represents an incoming profile of a roll pass."""

    class OutProfile(Profile, DiskedUnit.OutProfile):
        """Represents an outgoing profile of a roll pass."""

        filling_ratio = Hook[float]()

    class Roll(BaseRoll):
        """Represents a roll applied in a :py:class:`RollPass`."""

        def __init__(self, template: BaseRoll, roll_pass: 'RollPass'):
            kwargs = template.__dict__.copy()
            kwargs = dict([item for item in kwargs.items() if not item[0].startswith("_")])
            super().__init__(**kwargs)
            self.roll_pass = weakref.ref(roll_pass)

    class DiskElement(DiskedUnit.DiskElement):
        """Represents a disk element in a roll pass."""

        contact_area = Hook[float]()
        """Area of contact of the disk element to the rolls."""

        def roll_pass(self) -> 'RollPass':
            return self.parent()



import logging
import weakref
from typing import List

from ..hooks import Hook
from ..profile import Profile as BaseProfile
from ..disk_element import DiskedUnit


class Transport(DiskedUnit):
    """Represents a transport unit, e.g. an inter-rolling-stand gap, a furnace or cooling range."""

    environment_temperature = Hook[float]()
    """Temperature of the surrounding atmosphere."""

    def __init__(
            self,
            label: str = "",
            **kwargs
    ):
        """
        :param label: label for human identification
        :param kwargs: additional hook values as keyword arguments to set explicitly
        """

        super().__init__(label)
        self.__dict__.update(kwargs)
        self._log = logging.getLogger(__name__)

    @property
    def disk_elements(self) -> List['Transport.DiskElement']:
        """A list of disk elements used to subdivide this unit."""
        return list(self._subunits)

    class Profile(DiskedUnit.Profile):
        """Represents a profile in context of a transport unit."""

        def __init__(self, transport: 'Transport', template: BaseProfile):
            super().__init__(transport, template)
            self.transport = weakref.ref(transport)

    class InProfile(Profile, DiskedUnit.InProfile):
        """Represents an incoming profile of a transport unit."""

    class OutProfile(Profile, DiskedUnit.OutProfile):
        """Represents an outgoing profile of a transport unit."""

    class DiskElement(DiskedUnit.DiskElement):
        """Represents a disk element in a roll pass."""

        surface_area = Hook[float]()
        """Surface area of the disk element."""

        def transport(self) -> 'Transport':
            return self.parent()



from typing import Optional
import numpy as np
from scipy.optimize import minimize, Bounds
from numpy import sin, cos, tan, pi, array
from ..generic_elongation import GenericElongationGroove


class Oval3RadiiGroove(GenericElongationGroove):
    """Represents an oval shaped groove with 3 main radii."""

    def __init__(
            self,
            r1: float,
            r2: float,
            r3: float,
            depth: float,
            usable_width: float,
    ):
        """
        :param r1: radius of the first edge
        :type r1: float
        :param r2: radius of the second edge
        :type r2: float
        :param r2: radius of the third edge
        :type r2: float
        :param depth: depth of the groove
        :type depth: float
        :param usable_width:  ground width excluding influence of radii
        :type usable_width: float
        """

        r32 = r3 - r2

        def half_outer_width(alpha1):
            return usable_width / 2 + r1 * tan(alpha1 / 2)

        def geometric_conditions(x):
            alpha2 = x[0]
            alpha3 = x[1]
            alpha1 = alpha2 + alpha3
            gamma = pi / 2 - alpha2 - alpha3
            return array((
                r1 * sin(alpha1) + r2 * cos(gamma) + r32 * sin(alpha3) - half_outer_width(alpha1),
                r1 * (1 - cos(alpha1)) - r2 * sin(gamma) - r32 * cos(alpha3) + r3 - depth
            ))

        sol = minimize(lambda x: np.sum(geometric_conditions(x) ** 2, axis=0), array([pi / 4, pi / 4]),
                       bounds=Bounds(0, pi / 2), tol=1e-8 * depth)

        if not sol.success:
            raise RuntimeError("Could not determine geometric values with given input.")

        alpha2 = sol.x[0]
        alpha3 = sol.x[1]
        alpha1 = alpha2 + alpha3

        super().__init__(
            usable_width=usable_width, depth=depth,
            r1=r1, r2=r2, r3=r3,
            alpha1=alpha1, alpha2=alpha2, alpha3=alpha3
        )

    @property
    def types(self) -> '("oval", "oval_3_radii")':
        return "oval", "oval_3_radii"



import html
from abc import ABC, abstractmethod


class ReprMixin(ABC):
    """Mixin class providing common functionality for ``__repr__``, ``__str__``,
    ``_repr_pretty_`` and ``_repr_html_``."""

    @property
    @abstractmethod
    def __attrs__(self):
        raise NotImplementedError()

    def __str__(self):
        return type(self).__qualname__

    def __repr__(self):
        kwattrs = sorted(f"{name}={repr(value)}" for name, value in self.__attrs__.items())
        return f"{self.__class__.__name__}({', '.join(kwattrs)})"

    def _repr_pretty_(self, p, cycle):
        """Pretty printing for IPython."""

        if cycle:
            p.text(f"{type(self).__name__}(...)")
        else:
            with p.group(4, f"{type(self).__name__}(", ")"):
                p.break_()
                for name, value in sorted(self.__attrs__.items()):
                    p.text(name + "=")
                    p.pretty(value)
                    p.text(",")
                    p.breakable()

    # noinspection PyProtectedMember
    def _repr_html_(self):
        """HTML repr for IPython."""

        buf = [f"<table><tr><th colspan=2 style='text-align:center'>{html.escape(str(self), True)}</th></tr>"]

        for name, value in sorted(self.__attrs__.items()):
            if hasattr(value, "_repr_html_"):
                r = value._repr_html_()
            else:
                r = html.escape(repr(value), True)
            buf.append(f"<tr><td style='text-align:left'>{html.escape(name, True)}</td><td>{r}</td></tr>")

        buf.append("</table>")
        return ''.join(buf)



import logging
from typing import Union, Tuple

import numpy as np
from numpy import tan, sin, cos, pi, sqrt
from shapely.geometry import LineString, Polygon

from pyroll.core.grooves import GrooveBase
from pyroll.core.repr import ReprMixin


class GenericElongationGroove(GrooveBase, ReprMixin):
    """Represents a groove defined by the generic elongation groove geometry."""

    def __init__(
            self,
            usable_width: float = 0,
            depth: float = 0,

            r1: float = 0,
            r2: float = 0,
            r3: float = 0,
            r4: float = 0,

            even_ground_width: float = 0,
            indent: float = 0,

            alpha1: float = 0,
            alpha2: float = 0,
            alpha3: float = 0,
            alpha4: float = 0,

            types: Tuple[str, ...] = ()
    ):
        self.r1 = r1
        self.r2 = r2
        self.r3 = r3
        self.r4 = r4
        self.alpha1 = alpha1
        self.alpha2 = alpha2
        self.alpha3 = alpha3
        self.alpha4 = alpha4
        self.indent = indent
        self.even_ground_width = even_ground_width
        self._usable_width = usable_width
        self._depth = depth
        self._types = types

        self._log = logging.getLogger(__name__)

        self.z2 = usable_width / 2
        self.y2 = 0

        self.z1 = self.z2 + r1 * tan(alpha1 / 2)
        self.y1 = 0

        self.z12 = self.z1
        self.y12 = r1

        self.z3 = self.z1 - r1 * sin(alpha1)
        self.y3 = r1 * (1 - cos(alpha1))

        self.z9 = 0
        self.y9 = depth - indent

        self.z7 = even_ground_width / 2
        self.y7 = self.y9

        self.z8 = self.z7
        self.y8 = self.y9 + r4

        self.z6 = self.z8 + r4 * sin(alpha4)
        self.y6 = self.y8 - r4 * cos(alpha4)

        self.beta = alpha4 - alpha3 / 2

        self.z10 = self.z6 + r3 * sin(alpha3 / 2 + self.beta)
        self.y10 = self.y6 - r3 * cos(alpha3 / 2 + self.beta)

        self.z5 = self.z10 + r3 * sin(alpha3 / 2 - self.beta)
        self.y5 = self.y10 + r3 * cos(alpha3 / 2 - self.beta)

        self.z11 = self.z10 + (r3 - r2) * sin(alpha3 / 2 - self.beta)
        self.y11 = self.y10 + (r3 - r2) * cos(alpha3 / 2 - self.beta)

        self.gamma = pi / 2 - alpha2 - alpha3 + alpha4

        self.z4 = self.z11 + r2 * cos(self.gamma)
        self.y4 = self.y11 + r2 * sin(self.gamma)

        right_side = np.unique(list(self._enumerate_contour_points()), axis=0)
        left_side = np.flip(right_side, axis=0).copy()
        left_side[:, 0] *= -1

        self._contour_line = LineString(np.concatenate([left_side, right_side]))
        self._cross_section = Polygon(self._contour_line)

        self.test_plausibility()

    def _r1_contour_line(self, z):
        return self.y12 - sqrt(self.r1 ** 2 - (z - self.z12) ** 2)

    def _r2_contour_line(self, z):
        return self.y11 + sqrt(self.r2 ** 2 - (z - self.z11) ** 2)

    def _r3_contour_line(self, z):
        return self.y10 + sqrt(self.r3 ** 2 - (z - self.z10) ** 2)

    def _r4_contour_line(self, z):
        return self.y8 - sqrt(self.r4 ** 2 - (z - self.z8) ** 2)

    def _flank_contour_line(self, z):
        return self.y3 - tan(self.alpha1) * (z - self.z3)

    def _ground_contour_line(self, z):
        return np.ones_like(z) * (self.depth - self.indent)

    @staticmethod
    def _face_contour_line(z):
        return np.zeros_like(z)

    def _enumerate_contour_points(self):
        yield 1.1 * self.z1, self.y1

        for z in np.linspace(self.z1, self.z3, 20):
            yield z, self._r1_contour_line(z)

        for z in np.linspace(self.z4, self.z5, 20):
            yield z, self._r2_contour_line(z)

        if not np.isclose(self.z5, self.z6):
            for z in np.linspace(self.z5, self.z6, 20):
                yield z, self._r3_contour_line(z)

        if not np.isclose(self.z6, self.z7):
            for z in np.linspace(self.z6, self.z7, 20):
                yield z, self._r4_contour_line(z)

        yield self.z9, self.y9

    @property
    def contour_line(self) -> LineString:
        return self._contour_line

    @property
    def cross_section(self) -> Polygon:
        return self._cross_section

    def local_depth(self, z) -> Union[float, np.ndarray]:
        z = np.abs(z)

        return np.piecewise(
            z,
            [z < self.z7, (self.z7 <= z) & (z < self.z6), (self.z6 <= z) & (z < self.z5),
             (self.z5 <= z) & (z < self.z4), (self.z4 <= z) & (z < self.z3), (self.z3 <= z) & (z < self.z1)],
            [self._ground_contour_line, self._r4_contour_line, self._r3_contour_line, self._r2_contour_line,
             self._flank_contour_line, self._r1_contour_line, self._face_contour_line]
        )

    @property
    def types(self) -> Tuple[str, ...]:
        return self._types

    @property
    def usable_width(self) -> float:
        return self._usable_width

    @property
    def depth(self) -> float:
        return self._depth

    def test_plausibility(self):
        if (self.alpha1 + self.alpha4 - self.alpha2 - self.alpha3) > 0.01:
            raise ValueError("given angles should fulfill α1 + α4 = α2 + α3 to be geometrically plausible")

        if self.y4 - self._flank_contour_line(self.z4) > 0.001 * self.y4:
            print(self.y4 - self._flank_contour_line(self.z4), 0.1 * self.y4)
            raise ValueError("under given conditions a step appears in z4")

    @property
    def __attrs__(self):
        return {
            n: v for n in ["r1", "r2", "r3", "r4", "alpha1", "alpha2", "alpha3", "alpha4", "depth", "indent",
                           "even_ground_width", "usable_width", "types"]
            if (v := getattr(self, n))
        }



from typing import List

from ..unit import Unit
from ..hooks import Hook
from .disk_element import DiskElement as BaseDiskElement
from ..profile import Profile as BaseProfile


class DiskedUnit(Unit):
    """Base class for units that can be divided in disk elements."""

    disk_element_count = Hook[int]()
    """Number of intermediate profiles to create during unit solution (in and out profiles are not counted within)."""

    class Profile(Unit.Profile):
        """Represents a profile in context of a disked unit."""

    class InProfile(Profile, Unit.InProfile):
        """Represents an incoming profile of a disked unit."""

    class OutProfile(Profile, Unit.OutProfile):
        """Represents an outgoing profile of a disked unit."""

    class DiskElement(BaseDiskElement):
        """Represents a disk element in a roll pass."""

    @property
    def disk_elements(self) -> List['DiskElement']:
        """A list of disk elements used to subdivide this unit."""
        return self._subunits

    @property
    def __attrs__(self):
        return super().__attrs__ | {
            "disk_elements": self.disk_elements
        }

    def init_solve(self, in_profile: BaseProfile):
        super().init_solve(in_profile)
        self._subunits = self._SubUnitsList(self, [self.DiskElement(self, i) for i in range(self.disk_element_count)])



import logging
import weakref
from collections.abc import Sequence
from typing import overload, List

from ..unit import Unit
from ..profile import Profile as BaseProfile
from ..hooks import Hook


class PassSequence(Unit, Sequence[Unit]):
    """
    Represents a sequence of other units.
    Main container for defining rolling processes.
    Can be nested to define distinct rolling lines or similar.
    """

    total_elongation = Hook[float]()
    """Total elongation of the workpiece within the sequence."""

    def __init__(self, units: Sequence[Unit], label: str = "", **kwargs):
        """
        :param units: sequence of unit objects
        :param label: label for human identification
        :param kwargs: additional hook values as keyword arguments to set explicitly
        """

        super().__init__(label=label)
        self.__dict__.update(kwargs)
        self._subunits = self._SubUnitsList(self, units)
        self._log = logging.getLogger(__name__)

    class Profile(Unit.Profile):
        """Represents a profile in context of a pass sequence unit."""

        def __init__(self, pass_sequence: 'PassSequence', template: BaseProfile):
            super().__init__(pass_sequence, template)
            self.pass_sequence = weakref.ref(pass_sequence)

    class InProfile(Profile, Unit.InProfile):
        """Represents an incoming profile of a pass sequence unit."""

    class OutProfile(Profile, Unit.OutProfile):
        """Represents an outgoing profile of a pass sequence unit."""

    def __len__(self) -> int:
        return self._subunits.__len__()

    def __iter__(self):
        return self._subunits.__iter__()

    @overload
    def __getitem__(self, index: int) -> Unit:
        ...

    @overload
    def __getitem__(self, index: slice) -> list[Unit]:
        ...

    def __getitem__(self, index: int) -> Unit:
        return self._subunits.__getitem__(index)

    @property
    def units(self) -> List[Unit]:
        return self._subunits



import numpy as np

from .profile import Profile
from ..shapes import rectangle


@Profile.height
def height(self: Profile):
    if hasattr(self, "cross_section"):
        return np.abs(self.cross_section.bounds[3] - self.cross_section.bounds[1])


@Profile.width
def width(self: Profile):
    if hasattr(self, "cross_section"):
        return np.abs(self.cross_section.bounds[2] - self.cross_section.bounds[0])


@Profile.length
def length(self: Profile):
    return 0


@Profile.equivalent_height
def equivalent_height(self: Profile):
    return np.sqrt(self.cross_section.area * self.height / self.width)


@Profile.equivalent_height
def equivalent_height(self: Profile):
    if self.has_set_or_cached("equivalent_rectangle"):
        return self.equivalent_rectangle.height


@Profile.equivalent_width
def equivalent_width(self: Profile):
    return np.sqrt(self.cross_section.area * self.width / self.height)


@Profile.equivalent_width
def equivalent_width(self: Profile):
    if self.has_set_or_cached("equivalent_rectangle"):
        return self.equivalent_rectangle.width


@Profile.equivalent_rectangle
def equivalent_rectangle(self: Profile):
    return rectangle(self.equivalent_width, self.equivalent_height)


@Profile.surface_temperature
def surface_temperature(self: Profile):
    return self.temperature


@Profile.core_temperature
def core_temperature(self: Profile):
    return self.temperature



from typing import Optional
from scipy.optimize import brentq
from numpy import sin, cos, tan, pi

from ..generic_elongation import GenericElongationGroove


class Oval3RadiiFlankedGroove(GenericElongationGroove):
    """Represents an oval shaped groove with 3 main radii and a dedicated flank."""

    def __init__(
            self,
            r1: float,
            r2: float,
            r3: float,
            depth: float,
            usable_width: float,
            flank_angle: float,
    ):
        """
        Exactly two of ground_width, usable_width and flank_angle must be given.

        :param r1: radius of the first edge
        :type r1: float
        :param r2: radius of the second edge
        :type r2: float
        :param depth: depth of the groove
        :type depth: float
        :param usable_width:  ground width excluding influence of radii
        :type usable_width: float
        :param flank_angle: angle of the flanks
        :type flank_angle: float
        :raises ValueError:  if not exactly two of ground_width, usable_width and flank_angle are given
        """

        alpha1 = flank_angle
        r32 = r3 - r2
        gamma = pi / 2 - alpha1

        half_outer_width = usable_width / 2 + r1 * tan(alpha1 / 2)

        def geometric_conditions(alpha2):
            alpha3 = alpha1 - alpha2
            delta_w = r1 * sin(alpha1) + r2 * cos(gamma) + r32 * sin(alpha3) - half_outer_width
            delta_t = r1 * (1 - cos(alpha1)) - r2 * sin(gamma) - r32 * cos(alpha3) + r3 - depth
            flank_length = (delta_w - delta_t) / (sin(alpha1) - cos(alpha1))

            return delta_t + flank_length * sin(alpha1)

        try:
            sol = brentq(geometric_conditions, 0, alpha1)
        except:
            raise RuntimeError("Could not determine geometric values with given input.")

        alpha2 = sol
        alpha3 = alpha1 - alpha2

        super().__init__(
            usable_width=usable_width, depth=depth,
            r1=r1, r2=r2, r3=r3,
            alpha1=alpha1, alpha2=alpha2, alpha3=alpha3
        )

    @property
    def types(self) -> '("oval", "oval_3_radii", "oval_3_radii_flanked")':
        return "oval", "oval_3_radii", "oval_3_radii_flanked"



import numpy as np

from ..generic_elongation import GenericElongationGroove


class FalseRoundGroove(GenericElongationGroove):
    """Represents a round shaped groove with a dedicated flank (false round)."""

    def __init__(self, r1: float, r2: float, depth: float, flank_angle: float):
        """
        :param r1: radius of the first edge
        :type r1: float
        :param r2: radius of the second edge
        :type r2: float
        :param depth: depth of the groove
        :type depth: float
        :param flank_angle: angle of the flanks
        :type flank_angle: float
        """

        tip_depth = depth + r2 / np.cos(flank_angle) - r2
        usable_width = tip_depth / np.tan(flank_angle) * 2

        super().__init__(usable_width=usable_width, depth=depth, r1=r1, r2=r2, alpha1=flank_angle, alpha2=flank_angle)

    @property
    def types(self) -> '("round", "false_round")':
        return "round", "false_round"



from typing import Optional

import numpy as np

from ..generic_elongation import GenericElongationGroove


class ConstrictedBoxGroove(GenericElongationGroove):
    """Represents a box shaped groove with an indented ground."""

    def __init__(
            self,
            r1: float,
            r2: float,
            r4: float,
            depth: float,
            indent: float,
            ground_width: Optional[float] = None,
            usable_width: Optional[float] = None,
            flank_angle: Optional[float] = None
    ):
        """
        Exactly two of ground_width, usable_width and flank_angle must be given.

        :param r1: radius of the first edge
        :type r1: float
        :param r2: radius of the second edge
        :type r2: float
        :param depth: depth of the groove
        :type depth: float
        :param indent: indentation of the depth of the groove towards the grooves center
        :type indent: float
        :param ground_width: width of the groove from intersection between two flanks and ground width
        :type ground_width: float
        :param usable_width:  ground width excluding influence of radii
        :type usable_width: float
        :param flank_angle: angle of the flanks
        :type flank_angle: float
        :raises ValueError:  if not exactly two of ground_width, usable_width and flank_angle are given
        """
        if ground_width and usable_width and not flank_angle:
            flank_angle = np.arctan(depth / (usable_width - ground_width) * 2)
        elif usable_width and flank_angle and not ground_width:
            ground_width = usable_width - 2 * depth / np.tan(flank_angle)
        elif ground_width and flank_angle and not usable_width:
            usable_width = ground_width + 2 * depth / np.tan(flank_angle)
        else:
            raise ValueError(
                "Exactly two of the following arguments must be given: ground_width, usable_width, flank_angle must be given."
            )

        alpha4 = np.arccos(1 - indent / (r2 + r4))
        alpha2 = flank_angle + alpha4
        even_ground_width = ground_width - 2 * ((r4 + r2) * np.sin(alpha4) + r2 * np.tan(flank_angle / 2))

        super().__init__(
            usable_width=usable_width, depth=depth,
            r1=r1, r2=r2, r3=r2, r4=r4,
            alpha1=flank_angle, alpha2=alpha2, alpha4=alpha4,
            even_ground_width=even_ground_width, indent=indent
        )

    @property
    def types(self) -> '("box", "constricted_box")':
        return "box", "constricted_box"



import inspect
import logging
import weakref
from abc import ABCMeta
from functools import partial
from typing import overload, TypeVar, Generic, List, Generator, Union

import numpy as np

from pyroll.core.repr import ReprMixin

T = TypeVar("T")

_log = logging.getLogger(__name__)


class HookFunction:
    """
    Class wrapping a function used to yield the value of hooks.
    Instantiated commonly by use of a `Hook` instance as decorator on a function.
    You should not instantiate it yourself.
    """

    def __init__(self, func, hook, tryfirst=False, trylast=False):
        """
        :param func: the function
        :param hook: the associated hook
        :param tryfirst: whether to use this function with the highest priority
        :param trylast: whether to use this function with the lowest priority
        """
        self.function = func
        """The underlying function."""

        self.module = func.__module__
        """The module the function originates from."""

        self.qualname = func.__qualname__
        """The qualified name of the function."""

        self.name = func.__name__
        """The name of the function."""

        self.hook = hook
        """The hook the function is defined for."""

        self.cycle = False
        """Cycle detection."""

        self._tryfirst = tryfirst
        self._trylast = trylast

    @property
    def tryfirst(self):
        """Whether to use this function with the highest priority."""
        return self._tryfirst

    @property
    def trylast(self):
        """Whether to use this function with the lowest priority."""
        return self._trylast

    def __call__(self, instance):
        """Call the function as it were a method the provided instance."""

        extra_args = self._determine_extra_args()
        self.cycle = True
        try:
            result = self.function(instance, **extra_args)
        finally:
            self.cycle = False
        return result

    def __repr__(self):
        return f"<{self.__str__()}>"

    def __str__(self):
        return f"HookFunction {self.module}.{self.qualname}"

    def _determine_extra_args(self):
        extra_args = {}
        pars = inspect.signature(self.function).parameters
        if "cycle" in pars:
            extra_args["cycle"] = self.cycle

        return extra_args


class Hook(Generic[T]):
    """
    Descriptor yielding the value of a hook attribute if called on instance,
    or itself if called on class.
    """

    def __init__(self, name=None, owner=None):
        self.name = name
        """The name of the hook."""

        self.owner = owner
        """The owner class of the hook instance."""

        self._first_functions: List[HookFunction] = []

        self._last_functions: List[HookFunction] = []

        self._functions: List[HookFunction] = []

    def __set_name__(self, owner, name):
        self.name = name
        self.owner = owner

    @overload
    def __get__(self, instance: None, owner: type) -> 'Hook[T]':
        ...

    @overload
    def __get__(self, instance: 'HookHost', owner: type) -> T:
        ...

    def __get__(self, instance: 'HookHost', owner: type) -> Union[T, 'Hook[T]']:
        """
        Get the value of the hook if called on instance or the descriptor itself if called on class.

        Hook value resolution is done in the following order:
        1. explicit values (in ``__dict__``)
        2. cached values (in ``__cache__``)
        3. from hook functions.

        Saves the value in ``__cache__`` if the value was determined from functions.

        :raises AttributeError: if no value could be provided or if a RecursionError occurred during hook function calling
        :raises ValueError: if the resulting value was infinite (only for numeric values)
        """

        if self.owner != owner:
            # create distinct instance on subclass
            hook = Hook()
            setattr(owner, self.name, hook)
            return hook.__get__(instance, owner)

        if instance is None:
            return self

        # try to get value explicitly set by user
        result = instance.__dict__.get(self.name, None)
        if result is not None:
            return result

        # try to get cached value
        result = instance.__cache__.get(self.name, None)
        if result is not None:
            return result

        # try to get value from hook caller
        try:
            result = self.get_result(instance)
        except RecursionError as e:
            raise AttributeError(f"Hook call for '{self.name}' on '{instance}' resulted in a RecursionError. "
                                 f"This may have one of the following reasons: missing data, interference of plugins. "
                                 f"Double check if you have provided all necessary input data.") from e

        if result is None:
            raise AttributeError(f"Hook call for '{self.name}' on '{instance}' could not provide a value.")

        try:
            if not np.isfinite(result).all():
                raise ValueError(f"Hook call for '{self.name}' on '{instance}' resulted in an infinite value.")
        except TypeError:
            pass  # only numeric types can be tested for finiteness, for others it is meaningless

        instance.__cache__[self.name] = result

        return result

    def __set__(self, instance: object, value: T) -> None:
        """Saves a value to the ``__dict__`` of the instance."""
        instance.__dict__[self.name] = value

    def __delete__(self, instance: object) -> None:
        """
        Deletes the value from the ``__dict__`` of the instance.
        Does not raise, if the value is not in the ``__dict__``.
        """
        instance.__dict__.pop(self.name, None)

    def _yield_functions_from(self, attr: str):
        for s in self.owner.__mro__:
            h = getattr(s, self.name, None)
            funcs = getattr(h, attr, None)
            if funcs:
                yield from reversed(funcs)

    @property
    def functions_gen(self) -> Generator[HookFunction, None, None]:
        """
        Generator listing functions stored in this instance and equally named instances in superclasses of its owner.
        """

        yield from self._yield_functions_from("_first_functions")
        yield from self._yield_functions_from("_functions")
        yield from self._yield_functions_from("_last_functions")

    @property
    def functions(self) -> List[HookFunction]:
        """
        List of functions stored in this instance and equally named instances in superclasses of its owner.
        """
        return list(self.functions_gen)

    def get_result(self, instance) -> Union[T, None]:
        """
        Get the first not ``None`` result of the functions in ``self.functions`` or the cached value.
        """
        for f in self.functions_gen:
            result = f(instance)
            if result is not None:
                return result

    def add_function(self, func, tryfirst=False, trylast=False):
        """
        Add the given function to the internal function store.

        :return: the created HookFunction object
        """
        if isinstance(func, HookFunction):
            func = func.function

        hf = HookFunction(func, self, trylast=trylast, tryfirst=tryfirst)

        if tryfirst:
            self._first_functions.append(hf)
        elif trylast:
            self._last_functions.append(hf)
        else:
            self._functions.append(hf)

        return hf

    def __call__(self, func=None, tryfirst=False, trylast=False):
        if func is None:
            return partial(self.add_function, tryfirst=tryfirst, trylast=trylast)
        return self.add_function(func, tryfirst=tryfirst, trylast=trylast)

    def remove_function(self, func: HookFunction):
        """
        Remove a function from the internal function store.

        :return: the underlying function object
        """
        self._functions.remove(func)
        return func.function

    def __repr__(self):
        return f"<{self.__str__()}>"

    def __str__(self):
        return f"Hook {self.owner.__qualname__}.{self.name}"


class _HookHostMeta(ABCMeta):
    """
    Metaclass that provides plugin functionality to a class.

    Not for direct use but through :py:class:`PluginHost` base class.
    """

    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)

    def __setattr__(self, key, value):
        if isinstance(value, Hook):
            value.__set_name__(self, key)
        super().__setattr__(key, value)


class HookHost(ReprMixin, metaclass=_HookHostMeta):
    """
    A base class providing plugin functionality and some related convenience methods to a derived class.
    """

    def __init__(self):
        self.__cache__ = dict()

    def clear_hook_cache(self):
        """
        Clears the cache of hook function results.
        """
        self.__cache__.clear()

    def has_set(self, name: str):
        """Checks whether a value is explicitly set for the hook `name`."""
        return name in self.__dict__

    def has_cached(self, name: str):
        """Checks whether a value is cached for the hook `name`."""
        return name in self.__cache__

    def has_set_or_cached(self, name: str):
        """Checks whether a value is explicitly set or cached for the hook `name`."""
        return self.has_set(name) or self.has_cached(name)

    def has_value(self, name: str):
        """Checks whether a value is available for the hook `name`."""
        return hasattr(self, name)

    @classmethod
    @property
    def __hooks__(cls):
        """Return a set of all hook names defined on this class and all superclasses."""

        hooks = set()
        for s in cls.__mro__:
            hooks = hooks.union([
                name for name, value in s.__dict__.items()
                if not name.startswith("_")
                if isinstance(value, Hook)
            ])
        return hooks

    @classmethod
    def extension_class(cls, source: type):
        """
        Class decorator for adding new hook definitions to an existing HookHost.

        :return: the extended class (the reference to the source class is discarded)
        :raises TypeError: if the source class is itself derived from HookHost (error-prone)
        """

        for name, value in source.__dict__.items():
            if isinstance(value, Hook) and name not in cls.__dict__:
                setattr(cls, name, value)
        return cls

    @property
    def __attrs__(self):
        return {
            n: v
            for n, v in (self.__dict__ | self.__cache__).items()
            if not n.startswith("_") and not isinstance(v, weakref.ref)
        }

    def evaluate_and_set_hooks(self):
        """Evaluate functions of root hooks and set the results explicitly as attributes."""

        def _gen():
            for h in root_hooks:
                if issubclass(type(self), h.owner):
                    h = getattr(type(self), h.name)
                    result = h.get_result(self)

                    if result is None:
                        continue

                    setattr(self, h.name, result)

                    try:
                        arr = np.array(result)
                        yield from arr[np.isfinite(arr)].flat
                    except TypeError:  # yield only numeric values
                        continue

        return list(_gen())


root_hooks = set()  # filled in __init__.py due to circular imports
"""
Set of hooks to call explicitly in each solution iteration.
Their values will be treated as explicitly set but reevaluated in every iteration.
They will not be deleted during cache clearing.
They serve as root for the calling tree and persistent iterational variables.
"""
